@@ -16,6 +16,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
+#ident "$Id: iso2022.c,v 1.15 2002/12/10 00:08:32 nalin Exp $"
 #include "config.h"
 #include <sys/types.h>
 #include <stdio.h>
@@ -43,7 +44,7 @@ struct _vte_iso2022_map {
 
 struct _vte_iso2022 {
 	unsigned int current, override;
-	gboolean ss2, ss3;
+	gboolean ss0, ss1, ss2, ss3;
 	gunichar g[4];
 };
 
@@ -52,20 +53,35 @@ struct _vte_iso2022 {
 static const struct _vte_iso2022_map _vte_iso2022_map_0[] = {
 	{ 96, 0x25c6},	/* diamond */
 	{'a', 0x2592},	/* checkerboard */
+	{'b', 0x2409},	/* HT symbol */
+	{'c', 0x240c},	/* FF symbol */
+	{'d', 0x240d},	/* CR symbol */
+	{'e', 0x240a},	/* LF symbol */
 	{'f', GDK_degree},	/* degree */
 	{'g', GDK_plusminus},	/* plus/minus */
+	{'h', 0x2424},  /* NL symbol */
+	{'i', 0x240b},  /* VT symbol */
 	{'j', 0x2518},	/* downright corner */
 	{'k', 0x2510},	/* upright corner */
 	{'l', 0x250c},	/* upleft corner */
 	{'m', 0x2514},	/* downleft corner */
 	{'n', 0x253c},	/* cross */
-	{'q', 0x2500},	/* horizontal line */
+	{'o', 0x23ba},  /* scan line 1/9 */
+	{'p', 0x23bb},  /* scan line 3/9 */
+	{'q', 0x2500},	/* horizontal line (also scan line 5/9) */
+	{'r', 0x23bc},  /* scan line 7/9 */
+	{'s', 0x23bd},  /* scan line 9/9 */
 	{'t', 0x251c},	/* left t */
 	{'u', 0x2524},	/* right t */
-	{'v', 0x252c},	/* down t */
-	{'w', 0x2534},	/* up t */
+	{'v', 0x2534},	/* bottom t */
+	{'w', 0x252c},	/* top t */
 	{'x', 0x2502},	/* vertical line */
-	{127, 0x00b7},	/* bullet */
+	{'y', 0x2264},  /* <= */
+	{'z', 0x2265},  /* >= */
+	{'{', 0x03c0},  /* pi */
+	{'|', 0x2260},  /* not equal */
+	{'}', 0x00a3},  /* pound currency sign */
+	{'~', 0x00b7},	/* bullet */
 };
 /* United Kingdom.  VT100 and higher (per XTerm docs). */
 static const struct _vte_iso2022_map _vte_iso2022_map_A[] = {
@@ -76,133 +92,133 @@ static const struct _vte_iso2022_map _vte_iso2022_map_B[] = {
 };
 /* Dutch. VT220 and higher (per XTerm docs). */
 static const struct _vte_iso2022_map _vte_iso2022_map_4[] = {
-	{0x23, GDK_sterling},
-	{0x40, GDK_threequarters},
-	{0x5b, GDK_ydiaeresis},
-	{0x5c, GDK_onehalf},
-	{0x5d, GDK_bar},
-	{0x7b, GDK_diaeresis},
-	{0x7c, GDK_f},
-	{0x7d, GDK_onequarter},
-	{0x7e, GDK_acute}
+	{'#',  GDK_sterling},
+	{'@',  GDK_threequarters},
+	{'[',  GDK_ydiaeresis},
+	{'\\', GDK_onehalf},
+	{']',  GDK_bar}, /* FIXME? not in XTerm 170 */
+	{'{',  GDK_diaeresis},
+	{'|',  0x192}, /* f with hook (florin) */ /* FIXME? not in XTerm 170 */
+	{'}',  GDK_onequarter},
+	{'~',  GDK_acute}
 };
 /* Finnish. VT220 and higher (per XTerm docs). */
 static const struct _vte_iso2022_map _vte_iso2022_map_C[] = {
-	{0x5b, GDK_Adiaeresis},
-	{0x5c, GDK_Odiaeresis},
-	{0x5d, GDK_Aring},
-	{0x5e, GDK_Udiaeresis},
-	{0x60, GDK_eacute},
-	{0x7b, GDK_adiaeresis},
-	{0x7c, GDK_odiaeresis},
-	{0x7d, GDK_aring},
-	{0x7e, GDK_udiaeresis},
+	{'[',  GDK_Adiaeresis},
+	{'\\', GDK_Odiaeresis},
+	{']',  GDK_Aring},
+	{'^',  GDK_Udiaeresis},
+	{'`',  GDK_eacute},
+	{'{',  GDK_adiaeresis},
+	{'|',  GDK_odiaeresis},
+	{'}',  GDK_aring},
+	{'~',  GDK_udiaeresis},
 };
 /* French. VT220 and higher (per XTerm docs). */
 static const struct _vte_iso2022_map _vte_iso2022_map_R[] = {
-	{0x23, GDK_sterling},
-	{0x40, GDK_agrave},
-	{0x5b, GDK_degree},
-	{0x5c, GDK_ccedilla},
-	{0x5d, GDK_section},
-	{0x7b, GDK_eacute},
-	{0x7c, GDK_ugrave},
-	{0x7d, GDK_egrave},
-	{0x7e, GDK_diaeresis},
+	{'#',  GDK_sterling},
+	{'@',  GDK_agrave},
+	{'[',  GDK_degree},
+	{'\\', GDK_ccedilla},
+	{']',  GDK_section},
+	{'{',  GDK_eacute},
+	{'|',  GDK_ugrave},
+	{'}',  GDK_egrave},
+	{'~',  GDK_diaeresis},
 };
 /* French Canadian. VT220 and higher (per XTerm docs). */
 static const struct _vte_iso2022_map _vte_iso2022_map_Q[] = {
-	{0x40, GDK_agrave},
-	{0x5b, GDK_acircumflex},
-	{0x5c, GDK_ccedilla},
-	{0x5d, GDK_ecircumflex},
-	{0x5e, GDK_icircumflex},
-	{0x60, GDK_ocircumflex},
-	{0x7b, GDK_eacute},
-	{0x7c, GDK_ugrave},
-	{0x7d, GDK_egrave},
-	{0x7e, GDK_ucircumflex},
+	{'@',  GDK_agrave},
+	{'[',  GDK_acircumflex},
+	{'\\', GDK_ccedilla},
+	{']',  GDK_ecircumflex},
+	{'^',  GDK_icircumflex},
+	{'`',  GDK_ocircumflex},
+	{'{',  GDK_eacute},
+	{'|',  GDK_ugrave},
+	{'}',  GDK_egrave},
+	{'~',  GDK_ucircumflex},
 };
 /* German. VT220 and higher (per XTerm docs). */
 static const struct _vte_iso2022_map _vte_iso2022_map_K[] = {
-	{0x40, GDK_section},
-	{0x5b, GDK_Adiaeresis},
-	{0x5c, GDK_Odiaeresis},
-	{0x5d, GDK_Udiaeresis},
-	{0x7b, GDK_adiaeresis},
-	{0x7c, GDK_odiaeresis},
-	{0x7d, GDK_udiaeresis},
-	{0x7e, GDK_ssharp},
+	{'@',  GDK_section},
+	{'[',  GDK_Adiaeresis},
+	{'\\', GDK_Odiaeresis},
+	{']',  GDK_Udiaeresis},
+	{'{',  GDK_adiaeresis},
+	{'|',  GDK_odiaeresis},
+	{'}',  GDK_udiaeresis},
+	{'~',  GDK_ssharp},
 };
 /* Italian. VT220 and higher (per XTerm docs). */
 static const struct _vte_iso2022_map _vte_iso2022_map_Y[] = {
-	{0x23, GDK_sterling},
-	{0x40, GDK_section},
-	{0x5b, GDK_degree},
-	{0x5c, GDK_ccedilla},
-	{0x5d, GDK_eacute},
-	{0x60, GDK_ugrave},
-	{0x7b, GDK_agrave},
-	{0x7c, GDK_ograve},
-	{0x7d, GDK_egrave},
-	{0x7e, GDK_igrave},
+	{'#',  GDK_sterling},
+	{'@',  GDK_section},
+	{'[',  GDK_degree},
+	{'\\', GDK_ccedilla},
+	{']',  GDK_eacute},
+	{'`',  GDK_ugrave},
+	{'{',  GDK_agrave},
+	{'|',  GDK_ograve},
+	{'}',  GDK_egrave},
+	{'~',  GDK_igrave},
 };
 /* Norwegian and Danish. VT220 and higher (per XTerm docs). */
 static const struct _vte_iso2022_map _vte_iso2022_map_E[] = {
-	{0x40, GDK_Adiaeresis},
-	{0x5b, GDK_AE},
-	{0x5c, GDK_Ooblique},
-	{0x5d, GDK_Aring},
-	{0x5e, GDK_Udiaeresis},
-	{0x60, GDK_adiaeresis},
-	{0x7b, GDK_ae},
-	{0x7c, GDK_oslash},
-	{0x7d, GDK_aring},
-	{0x7e, GDK_udiaeresis},
+	{'@',  GDK_Adiaeresis},
+	{'[',  GDK_AE},
+	{'\\', GDK_Ooblique},
+	{']',  GDK_Aring},
+	{'^',  GDK_Udiaeresis},
+	{'`',  GDK_adiaeresis},
+	{'{',  GDK_ae},
+	{'|',  GDK_oslash},
+	{'}',  GDK_aring},
+	{'~',  GDK_udiaeresis},
 };
 /* Spanish. VT220 and higher (per XTerm docs). */
 static const struct _vte_iso2022_map _vte_iso2022_map_Z[] = {
-	{0x23, GDK_sterling},
-	{0x40, GDK_section},
-	{0x5b, GDK_exclamdown},
-	{0x5c, GDK_Ntilde},
-	{0x5d, GDK_questiondown},
-	{0x7b, GDK_degree},
-	{0x7c, GDK_ntilde},
-	{0x7d, GDK_ccedilla},
+	{'#',  GDK_sterling},
+	{'@',  GDK_section},
+	{'[',  GDK_exclamdown},
+	{'\\', GDK_Ntilde},
+	{']',  GDK_questiondown},
+	{'{',  GDK_degree},
+	{'|',  GDK_ntilde},
+	{'}',  GDK_ccedilla},
 };
 /* Swedish. VT220 and higher (per XTerm docs). */
 static const struct _vte_iso2022_map _vte_iso2022_map_H[] = {
-	{0x40, GDK_Eacute},
-	{0x5b, GDK_Adiaeresis},
-	{0x5c, GDK_Odiaeresis},
-	{0x5d, GDK_Aring},
-	{0x5e, GDK_Udiaeresis},
-	{0x60, GDK_eacute},
-	{0x7b, GDK_adiaeresis},
-	{0x7c, GDK_odiaeresis},
-	{0x7d, GDK_aring},
-	{0x7e, GDK_udiaeresis},
+	{'@',  GDK_Eacute},
+	{'[',  GDK_Adiaeresis},
+	{'\\', GDK_Odiaeresis},
+	{']',  GDK_Aring},
+	{'^',  GDK_Udiaeresis},
+	{'`',  GDK_eacute},
+	{'{',  GDK_adiaeresis},
+	{'|',  GDK_odiaeresis},
+	{'}',  GDK_aring},
+	{'~',  GDK_udiaeresis},
 };
 /* Swiss. VT220 and higher (per XTerm docs). */
 static const struct _vte_iso2022_map _vte_iso2022_map_equal[] = {
-	{0x23, GDK_ugrave},
-	{0x40, GDK_agrave},
-	{0x5b, GDK_eacute},
-	{0x5c, GDK_ccedilla},
-	{0x5d, GDK_ecircumflex},
-	{0x5e, GDK_icircumflex},
-	{0x5f, GDK_egrave},
-	{0x60, GDK_ocircumflex},
-	{0x7b, GDK_adiaeresis},
-	{0x7c, GDK_odiaeresis},
-	{0x7d, GDK_udiaeresis},
-	{0x7e, GDK_ucircumflex},
+	{'#',  GDK_ugrave},
+	{'@',  GDK_agrave},
+	{'[',  GDK_eacute},
+	{'\\', GDK_ccedilla},
+	{']',  GDK_ecircumflex},
+	{'^',  GDK_icircumflex},
+	{'_',  GDK_egrave},
+	{'`',  GDK_ocircumflex},
+	{'{',  GDK_adiaeresis},
+	{'|',  GDK_odiaeresis},
+	{'}',  GDK_udiaeresis},
+	{'~',  GDK_ucircumflex},
 };
 /* Japanese.  JIS X 0201-1976 ("Roman" set), per RFC 1468/2237. */
 static const struct _vte_iso2022_map _vte_iso2022_map_J[] = {
-	{'\\', GDK_overline},
-	{'~', GDK_yen},
+	{'\\', 0x203e},
+	{'~',  GDK_yen},
 };
 /* Japanese.  JIS X 0208-1978, per RFC 1468/2237. */
 static const struct _vte_iso2022_map _vte_iso2022_map_wide_at[] = {
@@ -240,6 +256,8 @@ _vte_iso2022_new(void)
 	ret = g_malloc0(sizeof(struct _vte_iso2022));
 	ret->current = 0;
 	ret->override = 0;
+	ret->ss0 = FALSE;
+	ret->ss1 = FALSE;
 	ret->ss2 = FALSE;
 	ret->ss3 = FALSE;
 	ret->g[0] = 'B';
@@ -320,6 +338,8 @@ _vte_iso2022_map_get(gunichar mapname)
 			ret = _vte_iso2022_map_init(_vte_iso2022_map_A,
 						    G_N_ELEMENTS(_vte_iso2022_map_A));
 			break;
+		case '1':
+		case '2':
 		case 'B':
 			ret = _vte_iso2022_map_init(_vte_iso2022_map_B,
 						    G_N_ELEMENTS(_vte_iso2022_map_B));
@@ -411,10 +431,27 @@ _vte_iso2022_map_get(gunichar mapname)
 	return ret;
 }
 
+gssize
+_vte_iso2022_get_width(gunichar c)
+{
+	gssize width;
+	width = (c & VTE_ISO2022_WIDTH_MASK) >> VTE_ISO2022_WIDTH_BIT_OFFSET;
+	return CLAMP(width, 0, 2);
+}
+
+static gunichar
+_vte_iso2022_set_width(gunichar c, gssize width)
+{
+	width = CLAMP(width, 0, 2);
+	c &= ~(VTE_ISO2022_WIDTH_MASK);
+	c |= (width << VTE_ISO2022_WIDTH_BIT_OFFSET);
+	return c;
+}
+
 gssize
 _vte_iso2022_substitute(struct _vte_iso2022 *outside_state,
 		       gunichar *instring, gssize length,
-		       gunichar *outstring, struct _vte_table *specials)
+		       gunichar *outstring, struct _vte_matcher *specials)
 {
 	int i, j, k, g;
 	struct _vte_iso2022 state;
@@ -425,6 +462,7 @@ _vte_iso2022_substitute(struct _vte_iso2022 *outside_state,
 	const char *match;
 	const gunichar *used;
 	int chars_per_code = 1;
+	gssize width;
 
 	g_return_val_if_fail(outside_state != NULL, 0);
 	g_return_val_if_fail(instring != NULL, 0);
@@ -436,8 +474,8 @@ _vte_iso2022_substitute(struct _vte_iso2022 *outside_state,
 
 	for (i = j = 0; i < length; i++) {
 		if ((specials != NULL) &&
-		    (_vte_table_match(specials, instring + i, length - i,
-				      &match, &used, NULL, NULL) != NULL)) {
+		    (_vte_matcher_match(specials, instring + i, length - i,
+					&match, &used, NULL, NULL) != NULL)) {
 			if (strlen(match) > 0) {
 				/* Aaargh, SI/SO masquerade as capabilities. */
 				if ((strcmp(match, "as") != 0) &&
@@ -703,6 +741,14 @@ _vte_iso2022_substitute(struct _vte_iso2022 *outside_state,
 			if (state.ss3) {
 				current_map = state.g[3];
 				state.ss3 = FALSE;
+			} else
+			if (state.ss0) {
+				current_map = state.g[0];
+				state.ss0 = FALSE;
+			} else
+			if (state.ss1) {
+				current_map = state.g[1];
+				state.ss1 = FALSE;
 			} else {
 				g_assert(state.current < G_N_ELEMENTS(state.g));
 				current_map = state.g[state.current];
@@ -787,6 +833,18 @@ _vte_iso2022_substitute(struct _vte_iso2022 *outside_state,
 #endif
 				}
 			}
+			/* Calculate width for ambiguous characters. */
+			switch (result) {
+				/* To be filled in some time, I guess. */
+			default:
+				/* For nonambiguous characters, just use
+				 * the width glib has later on. */
+				width = 0;
+				break;
+			}
+			if (width != 0) {
+				result = _vte_iso2022_set_width(result, width);
+			}
 			/* Store. */
 			buf[j++] = result;
 			accumulator = 0;
